#include "console.h"
#include "font.h"
#include "input.h"
#include "command.h"
#include "tcl_interp.h"
#include "tokenizer.h"

extern cmdlist cmdlst;

#define buf cmdlst.buf

decoration::decoration (float rr, float gg, float bb, int w) {
  r = rr;
  g = gg;
  b = bb;
  lw = w;
}

const float cc = 0.7;
const decoration console::light = decoration ();
const decoration console::yellow = decoration (1, 1, cc);
const decoration console::green = decoration (cc, 1, cc);
const decoration console::red = decoration (1, cc, cc);
const decoration console::cyan = decoration (cc, 1, 1);
const decoration console::bold = decoration (0, 0, 0, 2);

void console::push_back (const mesg& ln) {

  lines.push_back (ln);
  ++nlines;
  last ();
  
}

void console::up (int i) {
	startl = max (0, startl - i);
}

void console::down (int i) {
  startl += i;	
  if (startl >= nlines) startl = max (0, nlines - 1);
}

void console::last () {
  // ensures last line is always displayed
  end ();
  if (!rollup) up (lines_per_screen - 1);
}

void console::pgdn () {
  down (lines_per_screen);
}

void console::pgup () {
  up (lines_per_screen);
}

void console::home () {
  startl = 0;
}

void console::end () {
  startl = max (0, nlines - 1);
}

void console::add_last_line () {
	push_back (last_line);
	last_line = mesg ();
	loc = locx = 0;
}

console::console () {
  rollup = false;
  command_mode = 0;
	loc = 0;
	locx = 0;
  hid = -1;
	ps1 = "->";
  clear ();
}

console& console::operator<< (const string& s) {
	last_line.text += s;
	loc = last_line.text.length ();
	locx = x_after_string (last_line.text);
	return *this;
}

console& console::operator<< (unsigned int i) {
  sprintf (buf, "%x", i);
  *this << buf;
  return *this;
}

console& console::operator<< (int i) {
  sprintf (buf, "%d", i);
	*this << buf;
  return *this;
}

console& console::operator<< (unsigned long  i) {
  sprintf (buf, "%lu", i);
  *this << buf;
	return *this;
}

console& console::operator<< (unsigned long long i) {
  sprintf (buf, "%llux", i);
	*this << buf;
  return *this;
}

console& console::operator<< (float f) {
  sprintf (buf, "%f", f);
  *this << buf;
  return *this;
}

console& console::operator<< (double d) {
  sprintf (buf, "%f", d);
  *this << buf;
  return *this;
}

void console::clear_last_line () {
	last_line = mesg();
	loc = 0;
	locx = 0;
}

console& console::operator<< (char c) {

  switch (c) {

    case eol:
      add_last_line ();
      break;

    case cll:
			clear_last_line ();
      break;

    default:
			last_line.text.insert (last_line.text.begin()+loc, c);
			++loc; locx = x_after_string (last_line.text.substr (0, loc));
      break;
  }
  return *this;
}

void console::del () {
	int len = last_line.text.length ();
	if (loc > 0 && loc <= len) {
		last_line.text.erase (last_line.text.begin() + loc - 1);
		locx = (int) x_after_string (last_line.text.substr (0, --loc));
	}
}

console& console::operator<< (const decoration& d) {
  last_line.dec = d;
  return *this;
}

void console::calc_draw_params () {
	line_height = get_line_height ();
	char_width = max_char_width ();
  lines_per_screen = (win.top - win.bottom) / line_height - 2;
	startx = win.left + ps1.length () * char_width;
	starty = win.top - line_height;
}

void console::draw () {
  
  glMatrixMode (GL_MODELVIEW);
  
  glPushMatrix ();
  glLoadIdentity ();

  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();
  glOrtho (win.left, win.right, win.bottom, win.top, -1, 1);
    
  // draw lines
  int cx = startx, cy = starty;
	
  int n = 0, i = startl;
  while ((n++ < lines_per_screen) && (i < nlines)) {
    const mesg& linei = lines[i++];
    const decoration& dec = linei.dec;
    glColor3f (dec.r, dec.g, dec.b);
    draw_string (linei.text, cx, cy);
    cy -= line_height;
  }
	
  // draw last line
  glColor3f (1, 1, 1);
	draw_string (ps1, win.left, cy);
  draw_string (last_line.text, cx, cy, 0);
  
	// draw cursor
  if (command_mode) {
  	glColor3f (1, 0, 0);
  	int px = cx + locx, py = cy - 2;
  	glBegin (GL_LINES);
			glVertex2i (px, py);
			glVertex2i (px + char_width, py);
		glEnd ();
	}
  
  glPopMatrix ();

  glMatrixMode (GL_MODELVIEW);
  glPopMatrix ();


}

void console::clear () {
  lines.clear ();
  last_line = mesg();
  loc = 0; 
  locx = 0;
  nlines = 0;
  startl = 0;
}

void console::set_window (const box<int>& w) {
  win = w;
	calc_draw_params ();
}

bool console::handle_input () {
	
  if (command_mode) {
				
  	static const string cch [] = {
  		"abcdefghijklmnopqrstuvwxyz0123456789 .=-/;<\\,[]'`",
  		"ABCDEFGHIJKLMNOPQRSTUVWXYZ)!@#$%^&*( >+_?:<|<{}\"~"
  	};
  	static const char* cch0 = cch[0].c_str();
  	static const char* cch1 = cch[1].c_str();
		static const int j = cch[0].length ();
		
  	for (int i = 0; i < j; ++i) {
  		if (keypressedd (cch0[i])) {
  			if (keydown (SDLK_LSHIFT) || keydown (SDLK_RSHIFT) || keydown (SDLK_CAPSLOCK))
  				*this << cch1[i];
  			else
  				*this << cch0[i];
  			break;
  		}
		}

		if (keypressed (SDLK_RETURN)) {
			command = last_line.text;
			add_last_line ();
			int last = command.length () - 1;
			if ((last > 0) && (command[last] == ';')) command_mode = 1; else command_mode = 0;
			history.push_back (command);
			hid = history.size () - 1;
			extern tcl_interp interpreter;
			const char* result = interpreter (command);
			if (strlen (result)) {
				if (interpreter.result == TCL_ERROR) 
					*this << console::red << result << eol; 
				else 
					*this << console::green << result << eol;
			} 
		} else if (keypressedd (SDLK_LEFT)) {
			if (--loc < 1) { loc = 0; locx = 0;} else locx = (int) x_after_string (last_line.text.substr (0, loc));
		} else if (keypressedd (SDLK_RIGHT)) {
			int len = last_line.text.length ();
			if (++loc > len) loc = len;
			locx = (int) x_after_string (last_line.text.substr (0, loc));
		}
		
		if (hid > -1) {
			if (keypressedd (SDLK_UP)) {
				last_line.text = history [hid--]; 
				loc = last_line.text.length ();
				locx = x_after_string (last_line.text);
				if (hid < 0) hid = 0;
			}
			if (keypressedd (SDLK_DOWN)) {
				last_line.text = history [hid++];
				loc = last_line.text.length ();
				locx = x_after_string (last_line.text);
				int hs = history.size ();
				if (hid >= hs) hid = hs - 1;
			}
		}

	}
	
  static clock_t sc0 = secs2ticks(1./2), sc1 = secs2ticks(1./128);
  static clock_t pg0 = secs2ticks(1./2), pg1 = secs2ticks(1./16);
	if (keypressed (SDLK_TAB)) {
		loc = 0; locx = 0;
		last_line = mesg();
		command_mode = !command_mode;
		last ();
	} else	
  if (keypressedd (SDLK_INSERT, sc0, sc1)) up(1); else
  if (keypressedd (SDLK_DELETE, sc0, sc1)) down(1); else
  if (keypressed (SDLK_HOME)) home (); else
  if (keypressed (SDLK_END)) end(); else
  if (keypressedd (SDLK_PAGEUP, pg0, pg1)) pgup(); else
  if (keypressedd (SDLK_PAGEDOWN, pg0, pg1)) pgdn(); else
  if (keypressedd (SDLK_BACKSPACE)) {
		if (command_mode) del (); else clear();
  } else
  if (keypressed ('`') && !command_mode) {
		rollup = !rollup;
		last ();
	}
  return true;
  
}

console::~console () {}
